""" A view containing an interactive Python shell. """


# Standard library imports.
import logging
import traceback

# Major library imports
from IPython.kernel.core.interpreter import Interpreter

# Enthought library imports.
from envisage.api import IExtensionRegistry
from envisage.api import ExtensionPoint
from envisage.plugins.python_shell.api import IPythonShell
from envisage.plugins.ipython_shell.api import INamespaceView
from pyface.workbench.api import View
from pyface.ipython_widget import IPythonWidget
from pyface.api import GUI
from traits.api import Instance, Property, provides, Dict


# Setup a logger for this module.
logger = logging.getLogger(__name__)

@provides(IPythonShell)
class IPythonShellView(View):
    """ A view containing an IPython shell. """

    #### 'IView' interface ####################################################

    # The part's globally unique identifier.
    id = 'envisage.plugins.python_shell_view'

    # The part's name (displayed to the user).
    name = 'IPython'

    # The default position of the view relative to the item specified in the
    # 'relative_to' trait.
    position = 'bottom'

    #### 'PythonShellView' interface ##########################################

    # The interpreter's namespace.
    namespace = Dict

    # The names bound in the interpreter's namespace.
    names = Property(depends_on="namespace")

    #### 'IPythonShellView' interface #########################################

    # The interpreter
    interpreter = Instance(Interpreter)

    def _interpreter_default(self):
        # Create an interpreter that has a reference to our namespace.
        return Interpreter(user_ns=self.namespace)

    #### 'IExtensionPointUser' interface ######################################

    # The extension registry that the object's extension points are stored in.
    extension_registry = Property(Instance(IExtensionRegistry))

    #### Private interface ####################################################

    # Banner.
    _banner = ExtensionPoint(id='envisage.plugins.ipython_shell.banner')

    # Bindings.
    _bindings = ExtensionPoint(id='envisage.plugins.python_shell.bindings')

    # Commands.
    _commands = ExtensionPoint(id='envisage.plugins.python_shell.commands')

    ###########################################################################
    # 'IExtensionPointUser' interface.
    ###########################################################################

    def _get_extension_registry(self):
        """ Trait property getter. """

        return self.window.application

    ###########################################################################
    # 'View' interface.
    ###########################################################################

    def create_control(self, parent):
        """ Creates the toolkit-specific control that represents the view. """

        self.shell = IPythonWidget(parent,
                                   banner='\n'.join(self._banner),
                                   interp=self.interpreter)

        # Namespace contributions.
        for bindings in self._bindings:
            for name, value in bindings.items():
                self.bind(name, value)

        for command in self._commands:
            try:
                self.execute_command(command)
            except Exception as e:
                logger.exception(
                        "The command '%s' supplied to the Ipython shell "
                        "plugin has raised an exception:\n%s" %
                        (command, traceback.format_exc()))

        # Register the view as a service.
        self.window.application.register_service(IPythonShell, self)

        ns_view = self.window.application.get_service(INamespaceView)
        if ns_view is not None:
            self.on_trait_change(ns_view._on_names_changed, 'names')

        def try_set_focus():
            try:
                self.shell.control.SetFocus()
            except:
                # The window may not have been created yet.
                pass

        def set_focus():
            self.window.application.gui.invoke_later(try_set_focus)

        GUI.invoke_later(set_focus)

        return self.shell.control


    def destroy_control(self):
        """ Destroys the toolkit-specific control that represents the view.

        """

        super(IPythonShellView, self).destroy_control()

        # Remove the namespace change handler
        ns_view = self.window.application.get_service(INamespaceView)
        if ns_view is not None:
            self.on_trait_change(
                ns_view._on_names_changed, 'names', remove=True
            )


    ###########################################################################
    # 'PythonShellView' interface.
    ###########################################################################

    #### Properties ###########################################################

    def _get_names(self):
        """ Property getter. """

        return self.control.ipython0.magic_who_ls()

    #### Methods ##############################################################

    def bind(self, name, value):
        """ Binds a name to a value in the interpreter's namespace. """

        self.namespace[name] = value

        return

    def execute_command(self, command, hidden=True):
        """ Execute a command in the interpreter. """

        self.shell.execute_command(command, hidden)
        self.trait_property_changed('namespace', [], self.namespace)

    def execute_file(self, path, hidden=True):
        """ Execute a command in the interpreter. """

        self.shell.execute_file(path, hidden)
        self.trait_property_changed('namespace', [], self.namespace)

    def lookup(self, name):
        """ Returns the value bound to a name in the interpreter's namespace."""

        return self.namespace[name]

#### EOF ######################################################################



"""
A client controlling a remote editor.
"""

# Enthought library imports
from traits.api import provides

# Local imports
from envisage.plugins.remote_editor.communication.client import Client
from .i_remote_editor import IRemoteEditor


@provides(IRemoteEditor)
class RemoteEditorController(Client):
    """ A Client used to control a remote editor.
    """
    # Client interface

    self_type = "python_shell"
    other_type = "python_editor"

    def new_file(self):
        self.send_command('new')

    def open_file(self, filename):
        self.send_command('open', filename)





from .i_resource_protocol import IResourceProtocol
from .i_resource_manager import IResourceManager

from .file_resource_protocol import FileResourceProtocol
from .http_resource_protocol import HTTPResourceProtocol
from .no_such_resource_error import NoSuchResourceError
from .package_resource_protocol import PackageResourceProtocol
from .resource_manager import ResourceManager



""" A view showing a summary of the running application. """


# Standard library imports.
import inspect

# Enthought library imports.
from envisage.api import IApplication, IPlugin
from envisage.developer.code_browser.api import CodeBrowser
from apptools.io.api import File
from traits.api import Any, HasTraits, Instance
from traitsui.api import Item, TreeEditor, View

# fixme: non-api import.
from envisage.plugins.text_editor.editor.text_editor import TextEditor

# Local imports.
from .application_browser_tree import application_browser_tree_nodes


application_browser_view = View(
    Item(
        name       = 'application',
        show_label = False,
        editor     = TreeEditor(
            nodes       = application_browser_tree_nodes,
            editable    = False,
            orientation = 'vertical',
            hide_root   = True,
            show_icons  = True,
            selected    = 'selection',
            on_dclick   = 'object.dclick'
        )
    ),

    resizable = True,
    style     = 'custom',
    title     = 'Application',

    width     = .1,
    height    = .1
)


class ApplicationBrowser(HasTraits):
    """ An application browser.

    Actually, this class exists just because to use a trait editor we have
    to have a trait to edit!

    """

    # The application that we are browsing.
    application = Instance(IApplication)

    # The code browser that we use to parse plugin source code.
    code_browser = Instance(CodeBrowser)

    # The workbench service.
    workbench = Instance('envisage.ui.workbench.api.Workbench')

    # The object that is currently selected in the tree.
    selection = Any

    # The default traits UI view.
    traits_view = application_browser_view

    ###########################################################################
    # 'ApplicationBrowser' interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _code_browser_default(self):
        """ Trait initializer. """

        return self.application.get_service(CodeBrowser)

    def _workbench_default(self):
        """ Trait initializer. """

        from envisage.ui.workbench.api import Workbench

        return self.application.get_service(Workbench)

    #### Trait change handlers ################################################

    def _selection_changed(self, trait_name, old, new):
        """ Static trait change handler. """

        #print 'Selection changed', trait_name, old, new

        return

    #### Methods ##############################################################

    def dclick(self, obj):
        """ Called when an object in the tree is double-clicked. """

        if IPlugin(obj, None) is not None:
            # Parse the plugin source code.
            module = self._parse_plugin(obj)

            # Get the plugin klass.
            klass = self._get_plugin_klass(module, obj)

            # Edit the plugin.
            editor = self.workbench.edit(
                self._get_file_object(obj), kind=TextEditor
            )

            # Move to the class definition.
            editor.select_line(klass.lineno)

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_file_object(self, obj):
        """ Return a 'File' object for an object's source file. """

        return File(path=inspect.getsourcefile(type(obj)))

    def _get_plugin_klass(self, module, plugin):
        """ Get the klass that defines the plugin. """

        for name, klass in module.klasses.items():
            if name == type(plugin).__name__:
                break

        else:
            klass = None

        return klass

    def _parse_plugin(self, plugin):
        """ Parse the plugin source code. """

        filename = self._get_file_object(plugin).path

        return self.code_browser.read_file(filename)

#### EOF ######################################################################



""" Interface definition for the Remote shell """

# Enthought library imports.
from traits.api import Interface

class IRemoteShell(Interface):
    """ Interface definition for the remote shell """

    def run_file(self, path):
        """ Runs a file in the remote shell.

        """

    def run_text(self, text):
        """ Runs a string in the remote shell.

        """

#### EOF ######################################################################




""" Interface definition for the Remote editor """

# Enthought library imports.
from traits.api import Interface

class IRemoteEditor(Interface):
    """ Interface definition for the Namespace view """

    def new_file(self):
        """ Creates a new file in the remote editor.

        """

    def open_file(self, filename):
        """ Opens a file in the remote editor.

        """

#### EOF ######################################################################




"""Compatibility layer for Python version 2.x and 3.x.
"""

import sys

PY_VER = sys.version_info[0]

if PY_VER >= 3:
    import pickle
    from urllib.request import urlopen
    from urllib.error import HTTPError
    STRING_BASE_CLASS = str
else:
    import cPickle as pickle
    from urllib2 import urlopen, HTTPError
    STRING_BASE_CLASS = basestring

def unicode_str(x=''):
    return str(x) if PY_VER == 3 else unicode(x, encoding='utf-8')



""" Tests for the Enthought code browser. """

# Standard library imports.
import os.path
import logging

# Log to stidout (for now!).
logger = logging.getLogger()
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.DEBUG)


# Standard library imports.
import inspect, unittest

# Enthought library imports.
from traits.util.resource import get_path

from envisage.developer.code_browser.api import CodeBrowser



class CodeBrowserTestCase(unittest.TestCase):
    """ Tests for the Enthought code browser. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        self.code_browser = CodeBrowser()

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        return

    def assertIn(self, first, second, msg=None):
        # Python 2.6 compatibility layer.
        if hasattr(unittest.TestCase, 'assertIn'):
            unittest.TestCase.assertIn(self, first, second, msg)
        else:
            self.assertTrue(first in second, msg)

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_browse_this_module(self):
        """ browse this module """

        # We don't use '__file__' as the filename here as it could refer to the
        # '.pyc' file (we could just strip the 'c' of course, but inspect just
        # seems more intentional ;^).
        filename = inspect.getsourcefile(CodeBrowserTestCase)
        module = self.code_browser.read_file(filename)

        # Check the module name and documentation.
        self.assertTrue(module.name.endswith('code_browser_test_case'))
        self.assertEqual(" Tests for the Enthought code browser. ", module.doc)

        # Check that the module contains this class!
        klass = module.klasses['CodeBrowserTestCase']
        self.assertEqual(" Tests for the Enthought code browser. ", klass.doc)

        # And that the class contains this method.
        method = klass.methods['test_browse_this_module']
        self.assertEqual(" browse this module ", method.doc)

        return

    def test_has_traits(self):
        """ has traits """

        module = self.code_browser.read_file(
            os.path.join(get_path(CodeBrowserTestCase), 'example_1.py')
        )

        # Check the module name and documentation.
        self.assertTrue(module.name.endswith('example_1'))

        # Check that the module contains the specified class.
        klass = module.klasses.get('Base')
        self.assertNotEqual(None, klass)

        # Check the class' base class.
        self.assertEqual(['HasTraits'], klass.bases)

        # Check that the class has the appropriate traits and methods.
        self.assertEqual(2, len(klass.traits))
        self.assertIn("x", klass.traits)
        self.assertIn("y", klass.traits)

        self.assertEqual(2, len(klass.methods))
        self.assertIn("foo", klass.methods)
        self.assertIn("bar", klass.methods)

        return


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



""" Envisage package Copyright 2003-2007 Enthought, Inc. """

from .i_application import IApplication
from .i_extension_point import IExtensionPoint
from .i_extension_point_user import IExtensionPointUser
from .i_extension_provider import IExtensionProvider
from .i_extension_registry import IExtensionRegistry
from .i_import_manager import IImportManager
from .i_plugin import IPlugin
from .i_plugin_activator import IPluginActivator
from .i_plugin_manager import IPluginManager
from .i_service_registry import IServiceRegistry

from .application import Application
from .category import Category
from .class_load_hook import ClassLoadHook
from .egg_plugin_manager import EggPluginManager
from .extension_registry import ExtensionRegistry
from .extension_point import ExtensionPoint, contributes_to
from .extension_point_binding import ExtensionPointBinding, bind_extension_point
from .extension_provider import ExtensionProvider
from .extension_point_changed_event import ExtensionPointChangedEvent
from .import_manager import ImportManager
from .plugin import Plugin
from .plugin_activator import PluginActivator
from .plugin_extension_registry import PluginExtensionRegistry
from .plugin_manager import PluginManager
from .provider_extension_registry import ProviderExtensionRegistry
from .service import Service
from .service_offer import ServiceOffer
from .service_registry import NoSuchServiceError, ServiceRegistry
from .twisted_application import TwistedApplication
from .unknown_extension import UnknownExtension
from .unknown_extension_point import UnknownExtensionPoint


#### EOF ######################################################################



# Standard library imports
import os
import logging

# Local imports
from envisage.plugins.remote_editor.communication.client import Client


class EditorPlugin(Client):

    # EditorPlugin interface

    def new(self):
        raise NotImplementedError

    def open(self, filename):
        raise NotImplementedError

    def run_file(self, path):
        self.send_command('run_file', path)

    def run_text(self, text):
        self.send_command('run_text', text)

    # Client interface

    self_type = "python_editor"
    other_type = "python_shell"

    def handle_command(self, command, arguments):
        if command == "new":
            self.new()
            return True

        elif command == "open":
            if os.path.exists(arguments):
                self.open(arguments)
            else:
                logging.warning("EditorPlugin recieved invalid path '%s' for "
                                "'open' command", arguments)
            return True

        return False



# Enthought library imports.
from traits.api import HasTraits, Instance, Vetoable

# Local imports.
from .task_window import TaskWindow


class TaskWindowEvent(HasTraits):
    """ A task window lifecycle event.
    """

    # The window that the event occurred on.
    window = Instance(TaskWindow)


class VetoableTaskWindowEvent(TaskWindowEvent, Vetoable):
    """ A vetoable task window lifecycle event.
    """

    pass



""" A simple interface for the Python shell. """


# Enthought library imports.
from traits.api import Interface


class IPythonShell(Interface):
    """ A simple interface for the Python shell. """

    def bind(self, name, value):
        """ Binds a name to a value in the interpreter's namespace.

        """

    def execute_command(self, command, hidden=True):
        """ Execute a command in the interpreter.

        """

    def execute_file(self, path, hidden=True):
        """ Execute a file in the interpreter.

        """

    def lookup(self, name):
        """ Returns the value bound to a name in the interpreter's namespace.

        """

#### EOF ######################################################################



""" Envisage package Copyright 2003, 2004, 2005 Enthought, Inc. """


from .workbench import Workbench
from .workbench_action_set import WorkbenchActionSet
from .workbench_application import WorkbenchApplication
from .workbench_window import WorkbenchWindow


#### EOF ######################################################################



""" A definition of a category to be added to a class. """


# Enthought library imports.
from traits.api import HasTraits, Str


class Category(HasTraits):
    """ A definition of a category to be added to a class. """

    #### 'Category' interface #################################################

    # The name of the category class (the class that you want to add).
    class_name = Str

    # The name of the class that you want to add the category to.
    target_class_name = Str

#### EOF ######################################################################



"""
A client controlling a remote shell.
"""

# Enthought library imports
from traits.api import provides

# Local imports
from envisage.plugins.remote_editor.communication.client import Client
from .i_remote_shell import IRemoteShell


@provides(IRemoteShell)
class RemoteShellController(Client):
    """ A Client used to control a remote shell.
    """
    #------------------------------------------------------
    # Client interface
    #------------------------------------------------------

    self_type = "python_editor"
    other_type = "python_shell"

    #------------------------------------------------------
    # RemoteShell interface
    #------------------------------------------------------

    def run_file(self, path):
        self.send_command('run_file', path)

    def run_text(self, text):
        self.send_command('run_text', text)



""" The Enthought class browser module. """


# Standard library imports.
import imp, logging, os, stat, warnings

# Enthought library imports.
from apptools.io.api import File

# Local imports.
from .module import ModuleFactory
from .package import Package
from ..._compat import pickle

# Logging.
logger = logging.getLogger(__name__)


# Filter future warnings for maxint, since it causes warnings when compiling
# anything that has RGBA colors defined as hex values.
warnings.filterwarnings(
    "ignore", r'hex/oct constants > sys\.maxint .*', FutureWarning
)


# A cache that contains every module that has been parsed.
MODULES = {}

# Has the cache been modified in this process?
MODULES_CHANGED = False


def save_cache(filename):
    """ Saves the module cache to disk. """

    global MODULES
    global MODULES_CHANGED

    if MODULES_CHANGED:
        logger.debug('saving cache...')

        f = open(filename, 'wb')
        pickle.dump(MODULES, f, 1)
        f.close()

        logger.debug('cache saved')

    return


def load_cache(filename):
    """ Loads the module cache from disk. """

    global MODULES

    if os.path.isfile(filename):
        logger.debug('loading cache...')

        f = open(filename, 'rb')
        MODULES = pickle.load(f)
        f.close()

        logger.debug('cache loaded...')

    else:
        MODULES = {}

    return


def read_package(package_name):
    """ Parse every module in the specified package. """

    filename = find_module(package_name)
    if filename is None:
        raise ValueError("no such package '%s'" % package_name)

    package = Package(filename=filename, name=package_name)
    read_directory(filename, package)

    return package


def read_directory(filename, package=None):
    """ Parse every module in the specified directory. """

    directory = File(filename)
    if not directory.is_folder:
        raise ValueError("%s is NOT a directory." % filename)

    if package is not None:
        contents = package.contents

    else:
        contents = []

    for child in directory.children:
        if child.ext == '.py':
            contents.append(read_file(child.path, package))

        elif child.is_package:
            if package is not None:
                sub_package_name = '%s.%s' % (package.name, child.name)
                sub_package = Package(
                    filename=child.path, name=sub_package_name, parent=package
                )

            else:
                sub_package = Package(filename=child.path, name=child.name)

            read_directory(child.path, sub_package)
            contents.append(sub_package)

    return contents


def read_file(filename, namespace=None):
    """ Parses a file. """

    global MODULES
    global MODULES_CHANGED

    module, mod_time = MODULES.get(filename, (None, None))
    if module is None or mod_time != os.stat(filename)[stat.ST_MTIME]:
        logger.debug('parsing module %s' % filename)

        module_factory = ModuleFactory()
        try:
            module = module_factory.from_file(filename, namespace)

            # Add the parsed module to the cache.
            MODULES[filename] = (module, os.stat(filename)[stat.ST_MTIME])
            MODULES_CHANGED = True

        except:
            logger.exception('error parsing file %s' % filename)

    return module


def read_module(module_name):
    """ Parses a module. """

    filename = find_module(module_name)
    if filename is not None:
        module = read_file(filename)

    else:
        module = None

    return module


def find_module(module_name, path=None):
    """ Returns the filename for the specified module. """

    components = module_name.split('.')

    try:
        # Look up the first component of the module name (of course it could be
        # the *only* component).
        f, filename, description = imp.find_module(components[0], path)

        # If the module is in a package then go down each level in the package
        # hierarchy in turn.
        if len(components) > 0:
            for component in components[1:]:
                f,filename,description = imp.find_module(component, [filename])

    except ImportError:
        filename = None

    return filename

#### EOF ######################################################################



from envisage.ui.action.api import Action, ActionSet, Group
from pyface.action.api import Action as PyfaceAction
from envisage.plugins.python_shell.api import IPythonShell


def get_shell(window):
    """ Given an application window, retrieve the ipython shell.
    """
    return window.application.get_service(IPythonShell)

################################################################################
# Groups
################################################################################
ipython_shell_group = Group(
    id='IPythonShellGroup',
    path='MenuBar/Tools',
    #before='ExitGroup'
)


################################################################################
# `ClearScreen` class.
################################################################################
class ClearScreen(PyfaceAction):
    """ An action that clears the IPython screen. """

    tooltip      = "Clear the IPython screen."

    description  = "Clear the IPython screen."


    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Performs the action. """
        shell = get_shell(self.window)

        if shell is not None:
            shell.control.clear_screen()

clear_screen = Action(
    path        = "MenuBar/Tools",
    class_name  = __name__ + '.ClearScreen',
    name        = "Clear IPython screen",
    group       = "IPythonShellGroup",
)


################################################################################
# `IPythonShellActionSet` class.
################################################################################
class IPythonShellActionSet(ActionSet):
    """ The default action set for the IPython shell plugin. """

    groups = [ipython_shell_group, ]

    actions = [clear_screen]



""" Interface definition for the Namespace view """

# Enthought library imports.
from traits.api import Interface

class INamespaceView(Interface):
    """ Interface definition for the Namespace view """

    def _on_names_changed(self, new):
        """ Handler to track the changes in the namespace viewed.

        """

#### EOF ######################################################################



""" The Envisage core plugin. """


# Enthought library imports.
from envisage.api import ExtensionPoint, Plugin, ServiceOffer
from traits.api import List, Instance, on_trait_change, Str


class CorePlugin(Plugin):
    """ The Envisage core plugin.

    The core plugin offers facilities that are generally useful when building
    extensible applications such as adapters, categories and hooks etc. It does
    not contain anything to do with user interfaces!

    The core plugin should be started before any other plugin. It is up to
    the plugin manager to do this.

    """

    # Extension point Ids.
    CATEGORIES       = 'envisage.categories'
    CLASS_LOAD_HOOKS = 'envisage.class_load_hooks'
    PREFERENCES      = 'envisage.preferences'
    SERVICE_OFFERS   = 'envisage.service_offers'

    #### 'IPlugin' interface ##################################################

    # The plugin's unique identifier.
    id = 'envisage.core'

    # The plugin's name (suitable for displaying to the user).
    name = 'Core'

    #### Extension points offered by this plugin ##############################

    # Categories are actually implemented via standard 'ClassLoadHooks', but
    # for (hopefully) readability and convenience we have a specific extension
    # point.
    categories = ExtensionPoint(
        List(Instance('envisage.category.Category')),
        id   = CATEGORIES,
        desc = """

        Traits categories allow you to dynamically extend a Python class with
        extra attributes, methods and events.

        Contributions to this extension point allow you to import categories
        *lazily* when the class to be extended is imported or created. Each
        contribution contains the name of the category class that you want to
        add (the 'class_name') and the name of the class that you want to
        extend (the 'target_class_name').

        e.g. To add the 'FooCategory' category to the 'Foo' class::

            Category(
                class_name        = 'foo_category.FooCategory',
                target_class_name = 'foo.Foo'
            )

        """
    )
    @on_trait_change('categories_items')
    def _categories_items_changed(self, event):
        """ React to new categories being *added*.

        Note that we don't currently do anything if categories are *removed*.

        """

        self._add_category_class_load_hooks(event.added)

        return

    class_load_hooks = ExtensionPoint(
        List(Instance('envisage.class_load_hook.ClassLoadHook')),
        id   = CLASS_LOAD_HOOKS,
        desc = """

        Class load hooks allow you to be notified when any 'HasTraits' class
        is imported or created.

        See the documentation for 'ClassLoadHook' for more details.

        """
    )
    @on_trait_change('class_load_hooks_items')
    def _class_load_hooks_changed(self, event):
        """ React to new class load hooks being *added*.

        Note that we don't currently do anything if class load hooks are
        *removed*.

        """

        self._connect_class_load_hooks(event.added)

        return

    preferences = ExtensionPoint(
        List(Str),
        id   = PREFERENCES,
        desc = """

        Preferences files allow plugins to contribute default values for
        user preferences. Each contributed string must be the URL of a
        file-like object that contains preferences values.

        e.g.

        'pkgfile://envisage/preferences.ini'

        - this looks for the 'preferences.ini' file in the 'envisage'
        package.

        'file://C:/tmp/preferences.ini'

        - this looks for the 'preferences.ini' file in 'C:/tmp'

        'http://some.website/preferences.ini'

        - this looks for the 'preferences.ini' document on the 'some.website'
        web site!

        The files themselves are parsed using the excellent 'ConfigObj'
        package. For detailed documentation please go to:-

        http://www.voidspace.org.uk/python/configobj.html

        """
    )
    @on_trait_change('preferences_items')
    def _preferences_changed(self, event):
        """ React to new preferencess being *added*.

        Note that we don't currently do anything if preferences are *removed*.

        """

        self._load_preferences(event.added)

        return

    service_offers = ExtensionPoint(
        List(ServiceOffer),
        id   = SERVICE_OFFERS,
        desc = """

        Services are simply objects that a plugin wants to make available to
        other plugins. This extension point allows you to offer services
        that are created 'on-demand'.

        e.g.

        my_service_offer = ServiceOffer(
            protocol   = 'acme.IMyService',
            factory    = an_object_or_a_callable_that_creates_one,
            properties = {'a dictionary' : 'that is passed to the factory'}
        )

        See the documentation for 'ServiceOffer' for more details.

        """
    )
    @on_trait_change('service_offers_items')
    def _service_offers_changed(self, event):
        """ React to new service offers being *added*.

        Note that we don't currently do anything if services are *removed* as
        we have no facility to let users of the service know that the offer
        has been retracted.

        """
        for service in event.added:
            self._register_service_offer(service)

        return

    #### Contributions to extension points made by this plugin ################

    # None.

    ###########################################################################
    # 'IPlugin' interface.
    ###########################################################################

    def start(self):
        """ Start the plugin. """

        # Load all contributed preferences files into the application's root
        # preferences node.
        self._load_preferences(self.preferences)

        # Connect all class load hooks.
        self._connect_class_load_hooks(self.class_load_hooks)

        # Add class load hooks for all of the contributed categories. The
        # category will be imported and added when the associated target class
        # is imported/created.
        self._add_category_class_load_hooks(self.categories)

        # Register all service offers.
        #
        # These services are unregistered by the default plugin activation
        # strategy (due to the fact that we store the service ids in this
        # specific trait!).
        self._service_ids = self._register_service_offers(self.service_offers)

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _add_category_class_load_hooks(self, categories):
        """ Add class load hooks for a list of categories. """

        for category in categories:
            class_load_hook = self._create_category_class_load_hook(category)
            class_load_hook.connect()

        return

    def _connect_class_load_hooks(self, class_load_hooks):
        """ Connect all class load hooks. """

        for class_load_hook in class_load_hooks:
            class_load_hook.connect()

        return

    def _create_category_class_load_hook(self, category):
        """ Create a category class load hook. """

        # Local imports.
        from .class_load_hook import ClassLoadHook

        def import_and_add_category(cls):
            """ Import a category and add it to a class.

            This is a closure that binds 'self' and 'category'.

            """

            category_cls = self.application.import_symbol(category.class_name)
            cls.add_trait_category(category_cls)

            return

        category_class_load_hook = ClassLoadHook(
            class_name = category.target_class_name,
            on_load    = import_and_add_category
        )

        return category_class_load_hook

    def _load_preferences(self, preferences):
        """ Load all contributed preferences into a preferences node. """

        # Enthought library imports.
        from envisage.resource.api import ResourceManager

        # We add the plugin preferences to the default scope. The default scope
        # is a transient scope which means that (quite nicely ;^) we never
        # save the actual default plugin preference values. They will only get
        # saved if a value has been set in another (persistent) scope - which
        # is exactly what happens in the preferences UI.
        default = self.application.preferences.node('default/')

        # The resource manager is used to find the preferences files.
        resource_manager = ResourceManager()
        for resource_name in preferences:
            f = resource_manager.file(resource_name)
            try:
                default.load(f)

            finally:
                f.close()

        return

    def _register_service_offers(self, service_offers):
        """ Register a list of service offers. """

        return list(map(self._register_service_offer, service_offers))

    def _register_service_offer(self, service_offer):
        """ Register a service offer. """

        service_id = self.application.register_service(
            protocol   = service_offer.protocol,
            obj        = service_offer.factory,
            properties = service_offer.properties
        )

        return service_id

### EOF ######################################################################



""" The exception raised when an unknown extension is referenced. """


class UnknownExtension(Exception):
    """ The exception raised when an unknown extension is referenced. """


#### EOF ######################################################################



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2005-2007 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#
#-----------------------------------------------------------------------------

"""
A base class for actions that determine their enabled status based on
the state of the current project.

"""

# Standard library imports
import logging
import inspect

# Enthought library imports
from envisage.api import Application
from envisage.ui.workbench.api import WorkbenchWindow
from pyface.action.api import Action
from traits.api import Instance, Str

# Local imports.
from envisage.ui.single_project.services import IPROJECT_MODEL


# Setup a logger for this module.
logger = logging.getLogger(__name__)


class ProjectAction(Action):
    """
    A base class for actions that determine their enabled status based on
    the state of the current project.

    This class is architected such that implementors can override the
    'refresh' method without losing the default logic that determines
    the enabled state as a function of the current project.

    """
    #### 'ProjectAction' interface ####################################

    # The application that the action is part of. This is a convenience
    # property and is equivalent to 'self.window.application'.
    application = Instance(Application)

    # The project model service we refresh our state against.
    model_service = Instance(IPROJECT_MODEL)

    # The universal object locator (UOL). This string is used to locate an
    # instance to invoke a method on.
    #
    # UOLs can currently have one of the following forms:
    #
    # * ``'service://a_service_identifier'``
    # * ``'name://a/path/through/the/naming/service'``
    # * ``'file://the/pathname/of/a/file/containing/a/UOL'``
    # * ``'http://a/URL/pointing/to/a/text/document/containing/a/UOL'``
    uol = Str

    # The name of the method to invoke on the object.
    method_name = Str

    ##########################################################################
    # 'object' interface
    ##########################################################################

    def __init__(self, *args, **kws):
        """
        Constructor.

        Extended to setup a listener on the project model service so
        that we can refresh whenever the project changes.

        """

        # Retrieve the project model service and register ourself to listen
        # for project state changes.  This is done first to avoid errors
        # during any refresh calls triggered by further initialization.
        # FIXME: I don't think my implementation of the ProjectAction class is correct
        # because I can't see to get a reference to the current application.  Because of
        # this, I'm not able to setup the listeners yet but this needs to be done eventually!
        """
        if 'model_service' in kws:
            self.model_service = kws['model_service']
            del kws['model_service']
        else:
            self.model_service = self.window.application.get_service(IPROJECT_MODEL)
        self._update_model_service_listeners(remove=False)
        """

        super(ProjectAction, self).__init__(*args, **kws)

        return

    ##########################################################################
    # 'Action' interface
    ##########################################################################

    #### public interface ####################################################

    def destroy(self):
        """
        Destroy this action.

        Overridden here to remove our project model service listeners.

        """

        self._update_model_service_listeners(remove=True)

        return

    ##########################################################################
    # 'ProjectAction' interface
    ##########################################################################

    #### public interface ####################################################

    def refresh(self):
        """
        Refresh the enabled state of this action.

        This default implementation enables the action only when there is a
        current project.

        """

        self.enabled = self._refresh_project_exists()

        return

    def perform(self, event):
        """ Performs the action.

        This implementation simply performs the action specified by **uol**
        and **method_name**.

        Override this method to add additional work to the performance of this
        action.

        """

        self._perform_uol_action(event)

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _perform_uol_action(self, event):
        """ Called to perform the configured UOL action. """

        # Find the object.
        object = event.window.application.get_service(self.uol)
        if object is not None:
            method = getattr(object, self.method_name)

            # If the only argument is 'self' then don't pass the event. This
            # makes it easier to hook up actions that invoke NON-UI methods.
            #
            # fixme: Should we check the argument spec of the method more
            # closely and only pass the event iff there is exactly one argument
            # called 'event'?
            args, varargs, varkw, dflts = inspect.getargspec(method)
            if len(args) == 1:
                method()

            else:
                method(event)

        else:
            logger.error("Cannot resolve UOL: %s" % self.uol)

        return


    #### protected interface #################################################

    def _refresh_project_exists(self):
        """
        Return the refresh state according to whether the model service has a
        current project.

        Returns True if this action should be enabled.  False otherwise.

        """

        enabled = False
        if self.model_service is not None \
            and self.model_service.project is not None:
            enabled = True

        return enabled


    def _update_model_service_listeners(self, remove=False):
        """
        Update our listeners on the project model service.

        These are done as individual listener methods so that derived
        classes can change the behavior when a single event occurs.

        """

        logger.debug( (remove and 'Removing ' or 'Adding ') + \
            'listeners on project model service for ProjectAction [%s]', self)

        self.model_service.on_trait_change(self._on_project_changed,
            'project', remove=remove)
        self.model_service.on_trait_change(self._on_project_selection_changed,
            'selection', remove=remove)

        return


    #### trait handlers ######################################################

    def _on_project_changed(self, obj, trait_name, old, new):
        """
        Handle changes to the value of the current project.

        """

        self.refresh()

        return


    def _on_project_selection_changed(self, obj, trait_name, old, new):
        """
        Handle changes to the selection value within the current project.

        """

        self.refresh()

        return


#### EOF #####################################################################



# Copyright (c) 2007-2015 by Enthought, Inc.
# All rights reserved.

from ._version import full_version as __version__  # noqa

__requires__ = [
    'apptools',
    'traits',
]

# Per logging best practices, add a NullHandler to the root 'envisage'
# logger.
import logging


class NullHandler(logging.Handler):
    # Define our own to accommodate Python < 2.7
    def handle(self, record):
        pass

    def emit(self, record):
        pass

    def createLock(self):
        self.lock = None


logger = logging.getLogger(__name__)
logger.addHandler(NullHandler())

del logging, logger, NullHandler



"""
Envisage GUI Application
------------------------

This class handles the life-cycle of a Pyface GUI.  Plugins can
display windows via mechinisms such as edit_traits().

This is intended to be a very simple shell for lifting an existing
pure TraitsUI or Pyface (or even Qt) app into an Envisage app.

More sophisticated applications should use Tasks.

"""

from traits.api import Event, Supports
from envisage.api import Application


class GUIApplication(Application):
    """ The entry point for an Envisage GUI application.

    This class handles the life-cycle of a Pyface GUI.  Plugins can
    display windows via mechinisms such as edit_traits().
    
    This is intended to be a very simple shell for lifting an existing
    pure TraitsUI or Pyface (or even Qt) app into an Envisage app.
    
    More sophisticated applications should use Tasks.
    
    """

    #### 'GUIApplication' interface #########################################

    #: The PyFace GUI for the application.
    gui = Supports('pyface.i_gui.IGUI')

    #: The splash screen for the application. By default, there is no splash
    #: screen.
    splash_screen = Supports('pyface.i_splash_screen.ISplashScreen')

    #### Application lifecycle events #########################################

    #: Fired after the GUI event loop has been started.
    application_initialized = Event

    ###########################################################################
    # 'IApplication' interface.
    ###########################################################################

    def run(self):
        """ Run the application.

        Returns:
        --------
        Whether the application started successfully (i.e., without a veto).
        
        """
        
        # Make sure the GUI has been created (so that, if required, the splash
        # screen is shown).
        gui = self.gui

        started = self.start()
        if started:
            gui.set_trait_later(self, 'application_initialized', self)
            # Start the GUI event loop.  The application will block here.
            gui.start_event_loop()
            
            # clean up plugins once event loop stops
            self.stop()

        return started

    #### Trait initializers ###################################################

    def _gui_default(self):
        from pyface.api import GUI
        return GUI(splash_screen=self.splash_screen)




""" A trait type used to declare and access extension points. """


# Standard library imports.
import inspect, weakref

# Enthought library imports.
from traits.api import List, TraitType, Undefined, provides

# Local imports.
from .i_extension_point import IExtensionPoint



def contributes_to(id):
    """ A factory for extension point decorators!

    As an alternative to making contributions via traits, you can use this
    decorator to mark any method on a 'Plugin' as contributing to an extension
    point (note this is *only* used on 'Plugin' instances!).

    e.g. Using a trait you might have something like::

        class MyPlugin(Plugin):
            messages = List(contributes_to='acme.messages')

            def _messages_default(self):
                return ['Hello', 'Hola']

    whereas, using the decorator, it would be::

        class MyPlugin(Plugin):
            @contributes_to('acme.messages')
            def _get_messages(self):
                return ['Hello', 'Hola']

    There is not much in it really, but the decorator version looks a little
    less like 'magic' since it doesn't require the developer to know about
    Traits default initializers. However, if you know that you will want to
    dynamically change your contributions then use the trait version  because
    all you have to do is change the value of the trait and the framework will
    react accordingly.

    """

    def decorator(fn):
        """ A decorator for marking methods as extension contributors. """

        fn.__extension_point__ = id

        return fn

    return decorator


# Exception message template.
INVALID_TRAIT_TYPE = 'extension points must be "List"s e.g. List, List(Int)' \
' but a value of %s was specified.'


# Even though trait types do not themselves have traits, we can still
# declare that we implement an interface.
@provides(IExtensionPoint)
class ExtensionPoint(TraitType):
    """ A trait type used to declare and access extension points.

    Note that this is a trait *type* and hence does *NOT* have traits itself
    (i.e. it does *not* inherit from 'HasTraits').

    """

    ###########################################################################
    # 'ExtensionPoint' *CLASS* interface.
    ###########################################################################

    @staticmethod
    def bind(obj, trait_name, extension_point_id):
        """ Create a binding to an extension point. """

        from .extension_point_binding import bind_extension_point

        return bind_extension_point(obj, trait_name, extension_point_id)

    @staticmethod
    def connect_extension_point_traits(obj):
        """ Connect all of the 'ExtensionPoint' traits on an object. """

        for trait_name, trait in obj.traits(__extension_point__=True).items():
            trait.trait_type.connect(obj, trait_name)

        return

    @staticmethod
    def disconnect_extension_point_traits(obj):
        """ Disconnect all of the 'ExtensionPoint' traits on an object. """

        for trait_name, trait in obj.traits(__extension_point__=True).items():
            trait.trait_type.disconnect(obj, trait_name)

        return

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, trait_type=List, id=None, **metadata):
        """ Constructor. """

        # We add '__extension_point__' to the metadata to make the extension
        # point traits easier to find with the 'traits' and 'trait_names'
        # methods on 'HasTraits'.
        metadata['__extension_point__'] = True
        super(ExtensionPoint, self).__init__(**metadata)

        # The trait type that describes the extension point.
        #
        # If we are handed a trait type *class* e.g. List, instead of a trait
        # type *instance* e.g. List() or List(Int) etc, then we instantiate it.
        if inspect.isclass(trait_type):
            trait_type = trait_type()

        # Currently, we only support list extension points (we may in the
        # future want to allow other collections e.g. dictionaries etc).
        if not isinstance(trait_type, List):
            raise TypeError(INVALID_TRAIT_TYPE % trait_type)

        self.trait_type = trait_type

        # The Id of the extension point.
        if id is None:
            raise ValueError('an extension point must have an Id')

        self.id = id

        # A dictionary that is used solely to keep a reference to all extension
        # point listeners alive until their associated objects are garbage
        # collected.
        #
        # Dict(weakref.ref(Any), Dict(Str, Callable))
        self._obj_to_listeners_map = weakref.WeakKeyDictionary()

        return

    ###########################################################################
    # 'TraitType' interface.
    ###########################################################################

    def get(self, obj, trait_name):
        """ Trait type getter. """

        extension_registry = self._get_extension_registry(obj)

        # Get the extensions to this extension point.
        extensions = extension_registry.get_extensions(self.id)

        # Make sure the contributions are of the appropriate type.
        return self.trait_type.validate(obj, trait_name, extensions)

    def set(self, obj, name, value):
        """ Trait type setter. """

        extension_registry = self._get_extension_registry(obj)

        # Note that some extension registry implementations may not support the
        # setting of extension points (the default, plugin extension registry
        # for exxample ;^).
        extension_registry.set_extensions(self.id, value)

        return

    ###########################################################################
    # 'ExtensionPoint' interface.
    ###########################################################################

    def connect(self, obj, trait_name):
        """ Connect the extension point to a trait on an object.

        This allows the object to react when contributions are added or
        removed from the extension point.

        fixme: It would be nice to be able to make the connection automatically
        but we would need a slight tweak to traits to allow the trait type to
        be notified when a new instance that uses the trait type is created.

        """

        def listener(extension_registry, event):
            """ Listener called when an extension point is changed. """

            # If an index was specified then we fire an '_items' changed event.
            if event.index is not None:
                name = trait_name + '_items'
                old  = Undefined
                new  = event

            # Otherwise, we fire a normal trait changed event.
            else:
                name = trait_name
                old  = event.removed
                new  = event.added

            obj.trait_property_changed(name, old, new)

            return

        extension_registry = self._get_extension_registry(obj)

        # Add the listener to the extension registry.
        extension_registry.add_extension_point_listener(listener, self.id)

        # Save a reference to the listener so that it does not get garbage
        # collected until its associated object does.
        listeners = self._obj_to_listeners_map.setdefault(obj, {})
        listeners[trait_name] = listener

        return

    def disconnect(self, obj, trait_name):
        """ Disconnect the extension point from a trait on an object. """

        extension_registry = self._get_extension_registry(obj)

        listener = self._obj_to_listeners_map[obj].get(trait_name)
        if listener is not None:
            # Remove the listener from the extension registry.
            extension_registry.remove_extension_point_listener(
                listener, self.id
            )

            # Clean up.
            del self._obj_to_listeners_map[obj][trait_name]

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_extension_registry(self, obj):
        """ Return the extension registry in effect for an object. """

        extension_registry = getattr(obj, 'extension_registry', None)
        if extension_registry is None:
            # If the debug package is present then log the call stack to help
            # the developer see where the problem occurred.
            try:
                from etsdevtools.debug.api import log_called_from
                log_called_from(10)

            except:
                pass

            raise ValueError(
                'The "ExtensionPoint" trait type can only be used in ' \
                'objects that have a reference to an extension registry ' \
                'via their "extension_registry" trait. ' \
                'Extension point Id <%s>' % self.id
            )

        return extension_registry

#### EOF ######################################################################



""" The interface for objects using the 'Service' trait type. """


# Enthought library imports.
from traits.api import Instance, Interface

# Local imports.
from .i_service_registry import IServiceRegistry


class IServiceUser(Interface):
    """ The interface for objects using the 'Service' trait type. """

    # The service registry that the object's services are stored in.
    service_registry = Instance(IServiceRegistry)

#### EOF ######################################################################



""" Actions for the refresh code plugin. """


# Enthought library imports.
from pyface.action.api import Action


class RefreshCode(Action):
    """ Invoke the 'refresh code' function. """

    #### 'Action' interface ###################################################

    name        = 'Refresh Code'
    description = 'Refresh application to reflect python code changes'
    accelerator = 'Ctrl+Shift+R'

    def perform(self, event):
        """ Perform the action. """

        from traits.util.refresh import refresh

        refresh()

        return

#### EOF ######################################################################



""" A test class used in the service registry tests! """


# Enthought library imports.
from traits.api import HasTraits, provides

# Local imports.
from .i_foo import IFoo


@provides(IFoo)
class Foo(HasTraits):
    pass


#### EOF ######################################################################



""" A menu builder that doesn't build real actions! """


# Enthought library imports.
from envisage.ui.action.api import AbstractActionManagerBuilder
from pyface.action.api import Action, Group, MenuManager
from pyface.action.api import MenuBarManager
from traits.testing.unittest_tools import unittest


class DummyActionManagerBuilder(AbstractActionManagerBuilder):
    """ An action manager builder that doesn't build real actions!

    This makes it very easy to test!

    """

    ###########################################################################
    # 'DummyActionManagerBuilder' interface.
    ###########################################################################

    def create_menu_bar_manager(self, root):
        """ Create a menu bar manager from the builder's action sets. """

        menu_bar_manager = MenuBarManager(id='MenuBar')

        self.initialize_action_manager(menu_bar_manager, root)

        return menu_bar_manager

    ###########################################################################
    # Protected 'AbstractActionManagerBuilder' interface.
    ###########################################################################

    def _create_action(self, action_definition):
        """ Create an action implementation from a definition. """

        return Action(name=action_definition.class_name)

    def _create_group(self, group_definition):
        """ Create a group implementation from a definition. """

        return Group(id=group_definition.id)

    def _create_menu_manager(self, menu_definition):
        """ Create a menu manager implementation from a definition. """

        menu_manager = MenuManager(id=menu_definition.id)
        for group_definition in menu_definition.groups:
            menu_manager.insert(-1, Group(id=group_definition.id))

        return menu_manager

#### EOF ######################################################################



""" An action that opens a project. """

# Enthought library imports.
from envisage.ui.single_project.project_action import ProjectAction
from pyface.api import ImageResource

##############################################################################
# class 'OpenProjectAction'
##############################################################################

class OpenProjectAction(ProjectAction):
    """ An action that opens a project. """

    # The universal object locator (UOL).
    uol = 'envisage.ui.single_project.ui_service.UiService'

    # The name of the method to invoke on the object.
    method_name = 'open'

    # A longer description of the action.
    description = 'Open an existing project'

    # The action's image (displayed on tool bar tools etc).
    image = ImageResource('open_project')

    # The action's name (displayed on menus/tool bar tools etc).
    name = 'Open...'

    # A short description of the action used for tooltip text etc.
    tooltip = 'Open a project'



""" Tests for the plugin manager. """


# Enthought library imports.
from envisage.api import Plugin, PluginManager
from traits.api import Bool
from traits.testing.unittest_tools import unittest


class SimplePlugin(Plugin):
    """ A simple plugin. """

    #### 'SimplePlugin' interface #############################################

    started = Bool(False)
    stopped = Bool(False)

    ###########################################################################
    # 'IPlugin' interface.
    ###########################################################################

    def start(self):
        """ Start the plugin. """

        self.started = True
        self.stopped = False

        return

    def stop(self):
        """ Stop the plugin. """

        self.started = False
        self.stopped = True

        return


class BadPlugin(Plugin):
    """ A plugin that just causes trouble ;^). """

    ###########################################################################
    # 'IPlugin' interface.
    ###########################################################################

    def start(self):
        """ Start the plugin. """

        raise 1/0

    def stop(self):
        """ Stop the plugin. """

        raise 1/0


class PluginManagerTestCase(unittest.TestCase):
    """ Tests for the plugin manager. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_get_plugin(self):
        """ get plugin """

        simple_plugin  = SimplePlugin()
        plugin_manager = PluginManager(plugins=[simple_plugin])

        # Get the plugin.
        plugin = plugin_manager.get_plugin(simple_plugin.id)
        self.assertEqual(plugin, simple_plugin)

        # Try to get a non-existent plugin.
        self.assertEqual(None, plugin_manager.get_plugin('bogus'))

        return

    def test_iteration_over_plugins(self):
        """ iteration over plugins """

        simple_plugin = SimplePlugin()
        bad_plugin    = BadPlugin()

        plugin_manager = PluginManager(plugins=[simple_plugin, bad_plugin])

        # Iterate over the plugin manager's plugins.
        plugins = []
        for plugin in plugin_manager:
            plugins.append(plugin)

        self.assertEqual([simple_plugin, bad_plugin], plugins)

        return

    def test_start_and_stop(self):
        """ start and stop """

        simple_plugin  = SimplePlugin()
        plugin_manager = PluginManager(plugins=[simple_plugin])

        # Start the plugin manager. This starts all of the plugin manager's
        # plugins.
        plugin_manager.start()

        # Make sure the plugin was started.
        self.assertEqual(True, simple_plugin.started)

        # Stop the plugin manager. This stops all of the plugin manager's
        # plugins.
        plugin_manager.stop()

        # Make sure the plugin was stopped.
        self.assertEqual(True, simple_plugin.stopped)

        return

    def test_start_and_stop_errors(self):
        """ start and stop errors """

        simple_plugin  = SimplePlugin()
        bad_plugin     = BadPlugin()
        plugin_manager = PluginManager(plugins=[simple_plugin, bad_plugin])

        # Start the plugin manager. This starts all of the plugin manager's
        # plugins.
        self.failUnlessRaises(ZeroDivisionError, plugin_manager.start)

        # Stop the plugin manager. This stops all of the plugin manager's
        # plugins.
        self.failUnlessRaises(ZeroDivisionError, plugin_manager.stop)

        # Try to start a non-existent plugin.
        self.failUnlessRaises(
            SystemError, plugin_manager.start_plugin, plugin_id='bogus'
        )

        # Try to stop a non-existent plugin.
        self.failUnlessRaises(
            SystemError, plugin_manager.stop_plugin, plugin_id='bogus'
        )

        return

    def test_only_include_plugins_whose_ids_are_in_the_include_list(self):

        # Note that the items in the list use the 'fnmatch' syntax for matching
        # plugins Ids.
        include = ['foo', 'bar']

        plugin_manager = PluginManager(
            include = include,
            plugins = [
                SimplePlugin(id='foo'),
                SimplePlugin(id='bar'),
                SimplePlugin(id='baz')
            ]
        )

        # The Ids of the plugins that we expect the plugin manager to find.
        expected = ['foo', 'bar']

        # Make sure the plugin manager found only the required plugins and that
        # it starts and stops them correctly..
        self._test_start_and_stop(plugin_manager, expected)

        return

    def test_only_include_plugins_matching_a_wildcard_in_the_include_list(self):

        # Note that the items in the list use the 'fnmatch' syntax for matching
        # plugins Ids.
        include = ['b*']

        plugin_manager = PluginManager(
            include = include,
            plugins = [
                SimplePlugin(id='foo'),
                SimplePlugin(id='bar'),
                SimplePlugin(id='baz')
            ]
        )

        # The Ids of the plugins that we expect the plugin manager to find.
        expected = ['bar', 'baz']

        # Make sure the plugin manager found only the required plugins and that
        # it starts and stops them correctly..
        self._test_start_and_stop(plugin_manager, expected)

        return

    def test_ignore_plugins_whose_ids_are_in_the_exclude_list(self):

        # Note that the items in the list use the 'fnmatch' syntax for matching
        # plugins Ids.
        exclude = ['foo', 'baz']

        plugin_manager = PluginManager(
            exclude = exclude,
            plugins = [
                SimplePlugin(id='foo'),
                SimplePlugin(id='bar'),
                SimplePlugin(id='baz')
            ]
        )

        # The Ids of the plugins that we expect the plugin manager to find.
        expected = ['bar']

        # Make sure the plugin manager found only the required plugins and that
        # it starts and stops them correctly..
        self._test_start_and_stop(plugin_manager, expected)

        return

    def test_ignore_plugins_matching_a_wildcard_in_the_exclude_list(self):

        # Note that the items in the list use the 'fnmatch' syntax for matching
        # plugins Ids.
        exclude = ['b*']

        plugin_manager = PluginManager(
            exclude = exclude,
            plugins = [
                SimplePlugin(id='foo'),
                SimplePlugin(id='bar'),
                SimplePlugin(id='baz')
            ]
        )

        # The Ids of the plugins that we expect the plugin manager to find.
        expected = ['foo']

        # Make sure the plugin manager found only the required plugins and that
        # it starts and stops them correctly..
        self._test_start_and_stop(plugin_manager, expected)

        return

    #### Private protocol #####################################################

    def _test_start_and_stop(self, plugin_manager, expected):
        """ Make sure the plugin manager starts and stops the expected plugins.

        """

        # Make sure the plugin manager found only the required plugins.
        self.assertEqual(expected, [plugin.id for plugin in plugin_manager])

        # Start the plugin manager. This starts all of the plugin manager's
        # plugins.
        plugin_manager.start()

        # Make sure all of the the plugins were started.
        for id in expected:
            plugin = plugin_manager.get_plugin(id)
            self.assertNotEqual(None, plugin)
            self.assertEqual(True, plugin.started)

        # Stop the plugin manager. This stops all of the plugin manager's
        # plugins.
        plugin_manager.stop()

        # Make sure all of the the plugins were stopped.
        for id in expected:
            plugin = plugin_manager.get_plugin(id)
            self.assertNotEqual(None, plugin)
            self.assertEqual(True, plugin.stopped)

        return


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



""" Plugin finder for the current package. """


def get_plugins():
    """ Get the plugins from this package. """

    from .banana_plugin import BananaPlugin

    return [BananaPlugin()]

#### EOF #######################################################################



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2005, 2006 by Enthought, Inc.
#  All rights reserved.
#
#-----------------------------------------------------------------------------

"""
A base class for editors that can be tracked by single project plugin projects.

"""

# Standard library imports.
import logging

# Enthought library imports
from envisage.workbench import DecoratedEditor
from traits.api import Instance

# Application specific imports.
from envisage.single_project.services import IPROJECT_MODEL


# Setup a logger for this module.
logger=logging.getLogger(__name__)


class ProjectEditor(DecoratedEditor):
    """
    A base class for editors that can be tracked by single project plugin
    projects.

    """

    #########################################################################
    # Attributes
    #########################################################################

    ### public 'ProjectEditor' interface ####################################

    # The project containing the resource we're editing
    project = Instance('envisage.single_project.project.Project')


    #########################################################################
    # `object` interface
    #########################################################################

    #### operator methods ###################################################

    def __init__(self, **traits):
        """
        Constructor.

        Extended to associate ourself with the current project.

        """

        super(ProjectEditor, self).__init__(**traits)

        # Make sure the current project knows this editor is associated with
        # it's resources
        model_service = self.window.application.get_service(IPROJECT_MODEL)
        self.project = model_service.project
        self.project.register_editor(self.resource, self)

        return


    #########################################################################
    # 'Editor' interface.
    #########################################################################

    ### public 'Editor' interface ###########################################

    def destroy_control(self):
        """
        Destroys the toolkit-specific control that represents the editor.

        Extended to ensure that the current project stops associating us
        with its resources.

        """

        # Only do something if the editor is still open
        if self.control:
            logger.debug('Destroying control in ProjectEditor [%s]', self)

            # Unregister from the associated project immediately.
            self.project.register_editor(self.resource, self, remove=True)

        super(ProjectEditor, self).destroy_control()

        return


#### EOF ####################################################################
